/* copyright (C) 2001-2004 Jean-Luc Fontaine (mailto:jfontain@free.fr) */
/* this library is free software: please read the README file enclosed in this package */

/* $Id: tclperl.c,v 1.28 2006/02/26 18:57:39 jfontain Exp $ */

/*
    Provide Perl interpreters accessible from Tcl as a package named "tclperl".
    Also provides access to parent Tcl interpreter from Perl interpreters.
    See INSTALL and build.sh for build instructions.
*/

#include <tcl.h>
#include "tclperl.h"

#if (TCL_MAJOR_VERSION < 8)
    #error "Tcl 8.0 or greater is needed to build this"
#elif defined(USE_TCL_STUBS) && (TCL_MAJOR_VERSION == 8) &&\
    ((TCL_MINOR_VERSION == 0) || ((TCL_MINOR_VERSION == 1) && (TCL_RELEASE_LEVEL != TCL_FINAL_RELEASE)))
    #error "stubs interface does not work in 8.0 and alpha/beta 8.1, only 8.1.0+"
#endif

#ifdef _MSC_VER
    /* only do this when MSVC++ is used to compile */
    #ifdef USE_TCL_STUBS
        /* mark this .obj as needing Tcl stubs library */
        #pragma comment(lib, "tclstub" STRINGIFY(JOIN(TCL_MAJOR_VERSION,TCL_MINOR_VERSION)) ".lib")
        #if !defined(_MT) || !defined(_DLL) || defined(_DEBUG)
            /* fixes bug with how stubs library was compiled. requirement for msvcrt.lib from tclstubXX.lib should be removed. */
            #pragma comment(linker, "-nodefaultlib:msvcrt.lib")
        #endif
    #else
        /* mark this .obj as needing the import library */
        #pragma comment(lib, "tcl" STRINGIFY(JOIN(TCL_MAJOR_VERSION,TCL_MINOR_VERSION)) ".lib")
    #endif
#endif

/* to avoid warning as it is redefined in Tcl include file */
#undef STRINGIFY
#undef JOIN

#include <EXTERN.h>
#include <perl.h>
#ifdef MULTIPLICITY
extern void boot_DynaLoader(PerlInterpreter *, CV *);
#else
extern void boot_DynaLoader(CV *);
#endif


static struct Tcl_HashTable perlInterpreters;
#ifndef MULTIPLICITY
static unsigned existingInterpreters = 0;
#endif
static int newIdentifier = 0;
static Tcl_Interp *mainInterpreter;                                                   /* needed for Tcl evaluation from Perl side */
static Tcl_ThreadId mainThread;                                   /* needed for Perl threads, 0 if Tcl core is not thread-enabled */

extern void XS_Tcl_eval(PerlInterpreter *, CV *);
#ifdef MULTIPLICITY
static void xs_init(PerlInterpreter *my_perl)
#else
static void xs_init(void)
#endif
{                                                                                      /* allow dynamic loading of binary modules */
    newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, __FILE__);
    newXS("Tcl::eval", XS_Tcl_eval, __FILE__);
}

static int perlInterpreter(ClientData clientData, Tcl_Interp *interpreter, int numberOfArguments, Tcl_Obj * CONST arguments[])
{
    char *command;
    int identifier;
    struct Tcl_HashEntry *entry;
    PerlInterpreter *my_perl;
    SV *scalar;
    Tcl_Obj *object;

    if (numberOfArguments != 3) {
        Tcl_WrongNumArgs(interpreter, 1, arguments, "eval script");
        return TCL_ERROR;
    }
    command = Tcl_GetString(arguments[1]);
    if (strcmp(command, "eval") != 0) {
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(object, "bad option \"", command, "\": must be eval", 0);
        Tcl_SetObjResult(interpreter, object);
        return TCL_ERROR;
    }
    identifier = atoi(Tcl_GetString(arguments[0]) + 4);                                /* interpreter and command name is "perlN" */
    entry = Tcl_FindHashEntry(&perlInterpreters, (CONST char *)identifier);
    if (entry == 0) {
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(
            object, "invalid interpreter \"", Tcl_GetString(arguments[0]), "\": internal error, please report to tclperl author", 0
        );
        Tcl_SetObjResult(interpreter, object);
        return TCL_ERROR;
    }
    my_perl = (PerlInterpreter *)Tcl_GetHashValue(entry);
    PERL_SET_CONTEXT(my_perl);
    scalar = eval_pv(Tcl_GetString(arguments[2]), 0);                                                    /* catch errors manually */
    if (SvTRUE(ERRSV)) {
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(object, Tcl_GetString(arguments[0]), ": ", SvPV_nolen(ERRSV), 0); /* identify interpreter in error */
        Tcl_SetObjResult(interpreter, object);
        return TCL_ERROR;
    } else {
        Tcl_SetObjResult(interpreter, Tcl_NewStringObj(SvPV_nolen(scalar), -1));
        return TCL_OK;
    }
}

static int initializePerlInterpreter(PerlInterpreter *my_perl, Tcl_Interp *interpreter)
{
    CONST char *code = "{\
        package Tcl;\
        our $parent;\
        sub new($$) {\
            if (defined($parent)) {\
                die('fatal error: only 1 Tcl interpreter can exist at a time');\
            }\
            my $class = shift;\
            my $number = shift;\
            my $interpreter = 'tcl' . $number;\
            return bless(\\$interpreter, $class);\
        }\
    }";                                           /* note: taken from tcl.pl file, fully commented, used for testing in Perl only */
    Tcl_Obj *object;

    /* read(open("tcl.pl", O_RDONLY), code = malloc(100000), SSIZE_MAX); */                          /* used in development stage */
    eval_pv(code, 0);                                                                                    /* catch errors manually */
    if (!SvTRUE(ERRSV)) {
        eval_pv("$Tcl::parent = Tcl->new(0);", 0);
    }
    if (SvTRUE(ERRSV)) {
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(object, "fatal initialization error: ", SvPV_nolen(ERRSV), 0);
        Tcl_SetObjResult(interpreter, object);
        return TCL_ERROR;
    } else {
        return TCL_OK;
    }
}

static int newInterpreter(Tcl_Interp *interpreter)
{
    int identifier;
    char *embedded[] = {"", "-e", "0"};
    PerlInterpreter *my_perl;
    Tcl_Obj *object;
    int created;
    int result;

    my_perl = perl_alloc();
    PL_perl_destruct_level = 1;                                   /* make sure proper cleaning occurs when interpreter is deleted */
#ifndef MULTIPLICITY
    if (existingInterpreters > 0) {
        Tcl_SetResult(
            interpreter,
            "cannot create several concurrent Perl interpreters\n(Perl library was not compiled with -DMULTIPLICITY)",
            TCL_STATIC
        );
        return TCL_ERROR;
    }
#endif
    if (my_perl == 0) {
        Tcl_SetResult(interpreter, "fatal error: could not allocate a new Perl interpreter", TCL_STATIC);
        return TCL_ERROR;
    }
    PERL_SET_CONTEXT(my_perl);
    perl_construct(my_perl);
    perl_parse(my_perl, xs_init, 3, embedded, 0);
    perl_run(my_perl);
    result = initializePerlInterpreter(my_perl, interpreter);
    if (result != TCL_OK) {                                                 /* fatal initialization failure (should never happen) */
        perl_destruct(my_perl);
        perl_free(my_perl);
        return result;
    }
    identifier = newIdentifier;
    Tcl_SetHashValue(Tcl_CreateHashEntry(&perlInterpreters, (CONST char *)identifier, &created), my_perl);
    object = Tcl_NewStringObj("perl", -1);
    Tcl_AppendObjToObj(object, Tcl_NewIntObj(identifier));                            /* return "perlN" as interpreter identifier */
    Tcl_SetObjResult(interpreter, object);
    Tcl_CreateObjCommand(interpreter, Tcl_GetString(object), perlInterpreter, 0, 0);        /* create command for new interperter */
    newIdentifier++;
#ifndef MULTIPLICITY
    existingInterpreters++;
#endif
    return TCL_OK;
}

static int deleteInterpreters(Tcl_Interp *interpreter, int numberOfArguments, Tcl_Obj * CONST arguments[])
{
    int index;
    char *name;
    int identifier;
    struct Tcl_HashEntry *entry;
    PerlInterpreter *my_perl;

    for (index = 0; index < numberOfArguments; index++) {
        name = Tcl_GetString(arguments[index]);                                                    /* interpreter name is "perlN" */
        entry = 0;
        if (sscanf(name, "perl%u", &identifier) == 1) {
            identifier = atoi(name + 4);
            entry = Tcl_FindHashEntry(&perlInterpreters, (CONST char *)identifier);
        }
        if (entry == 0) {
            Tcl_Obj *object = Tcl_NewObj();
            Tcl_AppendStringsToObj(object, "invalid interpreter \"", name, 0);
            Tcl_SetObjResult(interpreter, object);
            return TCL_ERROR;
        }
        my_perl = (PerlInterpreter *)Tcl_GetHashValue(entry);
        PERL_SET_CONTEXT(my_perl);
        perl_destruct(my_perl);
        perl_free(my_perl);
        Tcl_DeleteHashEntry(entry);
#ifndef MULTIPLICITY
        existingInterpreters--;
#endif
    }
    return TCL_OK;
}

Tcl_Interp *tclInterpreter(CONST char *name)                           /* public function for use in extensions to this extension */
{
    int identifier;

    if ((sscanf(name, "tcl%u", &identifier) == 0) || (identifier != 0)) {
        return 0;                                                                                                 /* invalid name */
    } else {
        return mainInterpreter;                                                                     /* sole available interpreter */
    }
}

int tclEvaluate(CONST char *name, CONST char *script, char **string, int *length)                     /* returns true if no error */
{
    Tcl_Interp *interpreter;
    int result;
    Tcl_Obj *object;

    interpreter = tclInterpreter(name);
    if (interpreter == 0) {
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(object, "invalid Tcl interpreter name: ", name, 0);
        *string = Tcl_GetStringFromObj(object, length);
        return 0;
    } else if (CURRENTTHREAD == mainThread) {                                                                /* non threaded code */
        Tcl_Preserve(interpreter);
        result = Tcl_EvalEx(interpreter, script, -1, TCL_EVAL_DIRECT | TCL_EVAL_GLOBAL);
        *string = Tcl_GetStringFromObj(Tcl_GetObjResult(interpreter), length);
        Tcl_Release(interpreter);
        return (result == TCL_OK);
    } else {                               /* threaded code: function like Tcl thread extension send command in asynchronous mode */
        tclSendThread(mainThread, interpreter, script);               /* let the interpreter in the main thread evaluate the code */
        *string = 0;                                                                                       /* nothing is returned */
        return 1;                                                       /* errors, if any, are reported on standard error channel */
    }
}

static int command(ClientData clientData, Tcl_Interp *interpreter, int numberOfArguments, Tcl_Obj * CONST arguments[])
{
    char *command;
    unsigned new = 0;
    unsigned delete = 0;
    Tcl_Obj *object;

    if (numberOfArguments < 2) {
        Tcl_WrongNumArgs(interpreter, 1, arguments, "new|delete ?interp interp ...?");
        return TCL_ERROR;
    }
    command = Tcl_GetString(arguments[1]);
    new = (strcmp(command, "new") == 0);
    delete = (strcmp(command, "delete") == 0);
    if (!new && !delete) {
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(object, "bad option \"", command, "\": must be new or delete", 0);
        Tcl_SetObjResult(interpreter, object);
        return TCL_ERROR;
    }
    if (new) {
        if (numberOfArguments != 2) {
            Tcl_WrongNumArgs(interpreter, 1, arguments, "new");
            return TCL_ERROR;
        } else {
            return newInterpreter(interpreter);
        }
    }
    if (delete) {
        if (numberOfArguments < 3) {
            Tcl_WrongNumArgs(interpreter, 1, arguments, "delete ?interp interp ...?");
            return TCL_ERROR;
        } else {
            return deleteInterpreters(interpreter, numberOfArguments -2, arguments + 2);
        }
    }
    return TCL_ERROR;                                                                                            /* never reached */
}

#ifdef WIN32
/* George Petasis, 21 Feb 2006:
 * Under Visual C++, functions exported from DLLs must be declared
 * with __declspec(dllexport) */
#undef TCL_STORAGE_CLASS
#define TCL_STORAGE_CLASS DLLEXPORT
#endif

EXTERN int Tclperl_Init(Tcl_Interp *interpreter)
{
    static unsigned initialized = 0;
#ifdef _MSC_VER
    /* George Petasis, 21 Feb 2006:
     * Visual C++ does not seem to support the char *arguments[] = {};
     * syntax... */
    char **arguments = NULL;
#else
    char *arguments[] = {};
#endif
    int count;

#ifdef USE_TCL_STUBS
    if (Tcl_InitStubs(interpreter, "8.1", 0) == NULL) {
        return TCL_ERROR;
    }
#endif
    if (!initialized) {
        /* do the following once only since this function can be called several times if package is required from several Tcl */
        /* interpreters (thanks to Dave Bodenstab for discovering and fixing this bug).                                       */
        Tcl_InitHashTable(&perlInterpreters, TCL_ONE_WORD_KEYS);
        mainInterpreter = interpreter;
        mainThread = CURRENTTHREAD;
        count = sizeof(arguments) / sizeof(char *);                   /* avoid warnings by using variables, possibly unused below */
        PERL_SYS_INIT(&count, &arguments);                /* note: PERL_SYS_TERM() never used since extensions cannot be unloaded */
        initialized = 1;
    }
    Tcl_CreateObjCommand(interpreter, "::perl::interp", command, 0, 0);
    return Tcl_PkgProvide(interpreter, "tclperl", "3.2");
}
